# 2017 - Antonio Carlon

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from math import sqrt, sin, cos, pi

import resources
import threading
import qgis.utils

from qgis.core import *
from qgis.gui import *

name = "Split Features On Steroids"
moveVerticesName = "Move Vertices"
closeLineName = "Close line"
openLineName = "Open line"
moveLineName = "Move line"
areaUnits = {0 : "m<sup>2</sup>", 1 : "km<sup>2</sup>", 2 : "ft<sup>2</sup>", 3 : "sq yd", 4 : "sq mi", 5 : "ha", 6 : "ac", 7 : "M<sup>2</sup>", 8 : "deg<sup>2</sup>", 9 : ""}
maxDistanceHitTest = 5

class SplitFeaturesOnSteroidsPlugin:
	mapTool = None

	def __init__(self, iface):
		self.iface = iface

	def initGui(self):
		self.toolbar = self.iface.addToolBar(name)
		self.toolbar.setObjectName(name)
		
		icon = QIcon(":/plugins/SplitPolygonShowingAreas/icon.png")
		self.action = QAction(icon, name, self.iface.mainWindow())
		self.action.setCheckable(True)
		QObject.connect(self.action, SIGNAL("triggered()"), self.onClick)
		self.toolbar.addAction(self.action)

		self.actionMoveVertices = QAction(QIcon(":/plugins/SplitPolygonShowingAreas/moveVertices.png"), moveVerticesName, self.iface.mainWindow())
		self.actionMoveVertices.setCheckable(True)
		QObject.connect(self.actionMoveVertices, SIGNAL("triggered()"), self.onClickMoveVertices)
		self.toolbar.addAction(self.actionMoveVertices)

		self.actionLineClose = QAction(QIcon(":/plugins/SplitPolygonShowingAreas/lineClose.png"), closeLineName, self.iface.mainWindow())
		self.actionLineClose.setCheckable(False)
		QObject.connect(self.actionLineClose, SIGNAL("triggered()"), self.onClickLineClose)
		self.toolbar.addAction(self.actionLineClose)
		
		self.actionLineOpen = QAction(QIcon(":/plugins/SplitPolygonShowingAreas/lineOpen.png"), openLineName, self.iface.mainWindow())
		self.actionLineOpen.setCheckable(False)
		QObject.connect(self.actionLineOpen, SIGNAL("triggered()"), self.onClickLineOpen)
		self.toolbar.addAction(self.actionLineOpen)

		self.actionMoveLine = QAction(QIcon(":/plugins/SplitPolygonShowingAreas/moveLine.png"), moveLineName, self.iface.mainWindow())
		self.actionMoveLine.setCheckable(True)
		QObject.connect(self.actionMoveLine, SIGNAL("triggered()"), self.onClickMoveLine)
		self.toolbar.addAction(self.actionMoveLine)

		self.iface.addPluginToMenu(name, self.action)
		self.iface.addPluginToMenu(name, self.actionMoveVertices)
		self.iface.addPluginToMenu(name, self.actionLineClose)
		self.iface.addPluginToMenu(name, self.actionLineOpen)
		self.iface.addPluginToMenu(name, self.actionMoveLine)
		self.help_action = QAction("Help", self.iface.mainWindow())
		QObject.connect(self.help_action, SIGNAL("triggered()"), self.onHelp)
		self.iface.addPluginToMenu(name, self.help_action)

		self.iface.currentLayerChanged.connect(self.currentLayerChanged)

		self.enableTool()

	def disableAll(self):
		self.actionMoveVertices.setChecked(False)
		self.actionMoveVertices.setEnabled(False)
		self.actionLineClose.setEnabled(False)
		self.actionLineOpen.setEnabled(False)
		self.actionMoveLine.setChecked(False)
		self.actionMoveLine.setEnabled(False)

	def unload(self):
		self.iface.removePluginMenu(name, self.action)
		self.iface.removePluginMenu(name, self.actionMoveVertices)
		self.iface.removePluginMenu(name, self.actionLineClose)
		self.iface.removePluginMenu(name, self.actionLineOpen)
		self.iface.removePluginMenu(name, self.actionMoveLine)
		self.iface.removePluginMenu(name, self.help_action)
		self.iface.removeToolBarIcon(self.action)
		self.iface.removeToolBarIcon(self.actionMoveVertices)
		self.iface.removeToolBarIcon(self.actionLineClose)
		self.iface.removeToolBarIcon(self.actionLineOpen)
		self.iface.removeToolBarIcon(self.actionMoveLine)

	def onHelp(self):
		qgis.utils.showPluginHelp(filename="index")

	def onClick(self):
		self.disableAll()
		if not self.action.isChecked():
			if self.mapTool != None:
				self.mapTool.stopCapturing()
			self.iface.mapCanvas().unsetMapTool(self.mapTool)
			self.mapTool = None
			return
		layer = self.iface.activeLayer()
		if layer == None or not isinstance(layer, QgsVectorLayer) or (layer.wkbType() != QGis.WKBPolygon and layer.wkbType() != QGis.WKBMultiPolygon):
			self.iface.messageBar().pushMessage("No Polygon Vectorial Layer Selected", "Select a Polygon Vectorial Layer first", level=QgsMessageBar.WARNING)
			self.action.setChecked(False)
			return
		selectedFeatures = layer.selectedFeatures()
		if selectedFeatures == None or len(selectedFeatures) == 0:
			self.iface.messageBar().pushMessage("No Features Selected", "Select some features first", level=QgsMessageBar.WARNING)
			self.action.setChecked(False)
			return
		
		self.action.setChecked(True)
		self.mapTool = SplitMapTool(self.iface.mapCanvas(), layer, self.actionMoveVertices, self.actionLineClose, self.actionLineOpen, self.actionMoveLine)
		self.mapTool.setAction(self.action)
		self.iface.mapCanvas().setMapTool(self.mapTool)
		self.mapTool.redrawActions()

	def onClickMoveVertices(self):
		if not self.actionMoveVertices.isChecked():
			if self.mapTool != None:
				self.mapTool.stopMovingVertices()
			return

		self.mapTool.startMovingVertices()

	def onClickLineClose(self):
		self.mapTool.lineClose()

	def onClickLineOpen(self):
		self.mapTool.lineOpen()

	def onClickMoveLine(self):
		if not self.actionMoveLine.isChecked():
			if self.mapTool != None:
				self.mapTool.stopMovingLine()
			return

		self.actionMoveLine.setChecked(True)
		self.mapTool.startMovingLine()

	def currentLayerChanged(self):
		if self.mapTool != None:
			self.mapTool.stopCapturing()

		layer = self.iface.activeLayer()
		if layer != None:
			try:
				layer.editingStarted.disconnect(self.layerEditingChanged)
			except:
				pass
			try:
				layer.editingStopped.disconnect(self.layerEditingChanged)
			except:
				pass
			try:
				layer.selectionChanged .disconnect(self.layerEditingChanged)
			except:
				pass
			
			if isinstance(layer, QgsVectorLayer):
				layer.editingStarted.connect(self.layerEditingChanged)
				layer.editingStopped.connect(self.layerEditingChanged)
				layer.selectionChanged.connect(self.layerSelectionChanged)
			
		self.enableTool()		

	def layerEditingChanged(self):
		self.enableTool()	

	def layerSelectionChanged(self):
		self.enableTool()

	def enableTool(self):
		self.disableAll()
		self.action.setEnabled(False)
		layer = self.iface.activeLayer()
		
		if layer != None and isinstance(layer, QgsVectorLayer):
			selectedFeatures = layer.selectedFeatures()
			if isinstance(layer, QgsVectorLayer) and (layer.wkbType() == QGis.WKBPolygon or layer.wkbType() == QGis.WKBMultiPolygon) and selectedFeatures != None and len(selectedFeatures) > 0 and layer.isEditable():
				self.action.setEnabled(True)

class SplitMapTool(QgsMapToolEdit):
	def __init__(self, canvas, layer, actionMoveVertices, actionLineClose, actionLineOpen, actionMoveLine):
		super(SplitMapTool, self).__init__(canvas)
		self.canvas = canvas
		self.canvas.renderStarting.connect(self.mapCanvasChanged)
		self.canvas.keyReleased.connect(self.keyReleaseEvent)
		self.scene = canvas.scene()
		self.layer = layer
		self.layer.editingStopped.connect(self.stopCapturing)
		self.actionMoveVertices = actionMoveVertices
		self.actionLineClose = actionLineClose
		self.actionLineOpen = actionLineOpen
		self.actionMoveLine = actionMoveLine
		self.initialize()

	def initialize(self):
		self.selectedFeatures = self.layer.selectedFeatures()
		self.rubberBand = None
		self.tempRubberBand = None
		self.capturedPoints = []
		self.capturing = False
		self.setCursor(Qt.CrossCursor)
		self.proj = QgsProject.instance()
		self.labels = []
		self.vertices = []
		self.calculator = QgsDistanceArea()
		self.calculator.setSourceCrs(self.layer.dataProvider().crs())
		self.calculator.setEllipsoid(self.layer.dataProvider().crs().ellipsoidAcronym())
		self.calculator.setEllipsoidalMode(self.layer.dataProvider().crs().geographicFlag())
		self.drawingLine = False
		self.movingVertices = False
		self.movingLine = False
		self.showingVertices = False
		self.movingVertex = -1
		self.movingLineInitialPoint = None
		self.lineClosed = False

	def mapCanvasChanged(self):
		self.redrawAreas()
		if self.showingVertices:
			self.redrawVertices()

	def canvasMoveEvent(self, event):
		if self.drawingLine and not self.lineClosed:
			if self.tempRubberBand != None and self.capturing:
				mapPoint = self.toMapCoordinates(event.pos())
				self.tempRubberBand.movePoint(mapPoint)
				self.redrawAreas(event.pos())

		if self.movingVertices and self.movingVertex >= 0:
			layerPoint = self.toLayerCoordinates(self.layer, event.pos())
			self.capturedPoints[self.movingVertex] = layerPoint
			
			if self.lineClosed and self.movingVertex == 0:
				self.capturedPoints[len(self.capturedPoints) - 1] = layerPoint

			self.redrawRubberBand()
			self.redrawVertices()
			self.redrawAreas()

		if self.movingLine and self.movingLineInitialPoint != None:
			currentPoint = self.toLayerCoordinates(self.layer, event.pos())
			distance = self.distance(currentPoint, self.movingLineInitialPoint)
			bearing = self.movingLineInitialPoint.azimuth(currentPoint)
			for i in range(len(self.capturedPoints)):
				self.capturedPoints[i] = self.projectPoint(self.capturedPoints[i], distance, bearing)
			self.redrawRubberBand()
			self.redrawAreas()
			self.movingLineInitialPoint = currentPoint

	def projectPoint(self, point, distance, bearing):
		rads = bearing * pi / 180.0
   		dx = distance * sin(rads)
   		dy = distance * cos(rads)
   		return QgsPoint(point.x() + dx, point.y() + dy)
 	
	def redrawAreas(self, mousePos=None):
		self.deleteLabels()

		if self.capturing and len(self.capturedPoints) > 0:
			for i in range(len(self.selectedFeatures)):
				geometry = QgsGeometry(self.selectedFeatures[i].geometry())
				movingPoints = list(self.capturedPoints)
				
				if mousePos != None:
					movingPoints.append(self.toLayerCoordinates(self.layer, mousePos))

				result, newGeometries, topoTestPoints = geometry.splitGeometry(movingPoints, self.proj.topologicalEditing())

				self.addLabel(geometry)
				if newGeometries != None and len(newGeometries) > 0:
					for i in range(len(newGeometries)):
						self.addLabel(newGeometries[i])

	def addLabel(self, geometry):
		area = self.calculator.measureArea(geometry)
		labelPoint = geometry.pointOnSurface().vertexAt(0)
		label = QGraphicsTextItem("%.2f" % round(area,2))
		label.setHtml("<div style=\"color:#ffffff;background:#111111;padding:5px\">"
			+ "%.2f" % round(area,2) + " "
			+ areaUnits[self.calculator.areaUnits()]
			+ "</div>")
		label.setPos(self.toCanvasCoordinates(self.toMapCoordinates(self.layer, labelPoint)))

		self.scene.addItem(label)
		self.labels.append(label)

	def deleteLabels(self):
		for i in range(len(self.labels)):
			self.scene.removeItem(self.labels[i])
		self.labels = []

	def canvasPressEvent(self, event):
		if self.movingVertices:
			for i in range(len(self.capturedPoints)):
				currentVertex = self.toCanvasCoordinates(self.toMapCoordinates(self.layer, self.capturedPoints[i]))
				if self.distance(event.pos(), currentVertex) <= maxDistanceHitTest:
					self.movingVertex = i
					break
		elif self.movingLine:
			self.movingLineInitialPoint = self.toLayerCoordinates(self.layer, event.pos())

	def distance(self, eventPos, vertexPos):
		return sqrt((eventPos.x() - vertexPos.x())**2 + (eventPos.y() - vertexPos.y())**2)

	def canvasReleaseEvent(self, event):
		if self.movingVertices or self.movingLine:
			if event.button() == Qt.RightButton:
				self.finishOperation()
		else:
			if event.button() == Qt.LeftButton:
				if not self.lineClosed:
					if not self.capturing:
						self.startCapturing()
					self.addVertex(event.pos())
			elif event.button() == Qt.RightButton:
				self.finishOperation()

		self.movingVertex = -1
		self.movingLineInitialPoint = None
		self.redrawActions()

	def keyReleaseEvent(self, event):
		if event.key() == Qt.Key_Escape:
			self.stopCapturing()
		if event.key() == Qt.Key_Backspace or event.key() == Qt.Key_Delete:
			self.removeLastVertex()
			event.ignore()
		if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
			finishOperation()

		self.redrawActions()

	def finishOperation(self):
		self.doSplit()
		self.stopCapturing()
		self.initialize()
		self.startCapturing()

	def doSplit(self):
		if self.capturedPoints != None:
			self.layer.splitFeatures(self.capturedPoints, self.proj.topologicalEditing())

	def startCapturing(self):
		self.prepareRubberBand()
		self.prepareTempRubberBand()

		self.drawingLine = True
		self.capturing = True

		self.redrawActions()

	def prepareRubberBand(self):
		color = QColor("red")
		color.setAlphaF(0.78)

		self.rubberBand = QgsRubberBand(self.canvas, QGis.Line)
		self.rubberBand.setWidth(1)
		self.rubberBand.setColor(color)
		self.rubberBand.show()

	def prepareTempRubberBand(self):
		color = QColor("red")
		color.setAlphaF(0.78)

		self.tempRubberBand = QgsRubberBand(self.canvas, QGis.Line)
		self.tempRubberBand.setWidth(1)
		self.tempRubberBand.setColor(color)
		self.tempRubberBand.setLineStyle(Qt.DotLine)
		self.tempRubberBand.show()

	def redrawRubberBand(self):
		self.canvas.scene().removeItem(self.rubberBand)
		self.prepareRubberBand()
		for i in range(len(self.capturedPoints)):
			vertexCoord = self.toMapCoordinates(self.layer, self.capturedPoints[i])
			self.rubberBand.addPoint(vertexCoord)

	def redrawTempRubberBand(self):
		if self.tempRubberBand != None:
			self.tempRubberBand.reset(QGis.Line)
			self.tempRubberBand.addPoint(self.toMapCoordinates(self.layer, self.capturedPoints[len(self.capturedPoints) - 1]))

	def stopCapturing(self):
		self.deleteLabels()
		self.deleteVertices()
		if self.rubberBand:
			self.canvas.scene().removeItem(self.rubberBand)
			self.rubberBand = None
		if self.tempRubberBand:
			self.canvas.scene().removeItem(self.tempRubberBand)
			self.tempRubberBand = None
		self.drawingLine = False
		self.movingVertices = False
		self.showingVertices = False
		self.capturing = False
		self.capturedPoints = []
		self.canvas.refresh()

		self.redrawActions()

	def addVertex(self, canvasPoint):
		mapPoint = self.toMapCoordinates(canvasPoint)
		layerPoint = self.toLayerCoordinates(self.layer, canvasPoint)

		self.rubberBand.addPoint(mapPoint)
		self.capturedPoints.append(layerPoint)

		self.tempRubberBand.reset(QGis.Line)
		self.tempRubberBand.addPoint(mapPoint)

	def removeLastVertex(self):
		if not self.capturing: return

		rubberBandSize = self.rubberBand.numberOfVertices()
		tempRubberBandSize = self.tempRubberBand.numberOfVertices()
		numPoints = len(self.capturedPoints)

		if rubberBandSize < 1 or numPoints < 1:
			return

		self.rubberBand.removePoint(-1)

		if rubberBandSize > 1:
			if tempRubberBandSize > 1:
				point = self.rubberBand.getPoint(0, rubberBandSize-2)
				self.tempRubberBand.movePoint(tempRubberBandSize-2, point)
		else:
			self.tempRubberBand.reset(self.bandType())

		del self.capturedPoints[-1]

	def startMovingVertices(self):
		self.stopMovingLine()

		self.actionMoveVertices.setChecked(True)
		self.movingVertices = True
		self.showingVertices = True
		self.drawingLine = False
		self.canvas.scene().removeItem(self.tempRubberBand)
		self.redrawVertices()
		self.redrawAreas()
		self.redrawActions()

	def stopMovingVertices(self):
		self.actionMoveVertices.setChecked(False)
		self.movingVertices = False
		self.showingVertices = False
		self.drawingLine = True
		self.deleteVertices()
		self.redrawRubberBand()
		self.redrawTempRubberBand()
		self.canvas.scene().addItem(self.tempRubberBand)
		self.redrawActions()

	def startMovingLine(self):
		self.stopMovingVertices()

		self.actionMoveLine.setChecked(True)
		self.movingLine = True
		self.showingVertices = False
		self.drawingLine = False
		self.canvas.scene().removeItem(self.tempRubberBand)
		self.redrawAreas()
		self.redrawActions()

	def stopMovingLine(self):
		self.actionMoveLine.setChecked(False)
		self.movingLine = False
		self.showingVertices = False
		self.drawingLine = True
		self.redrawRubberBand()
		self.redrawTempRubberBand()
		self.canvas.scene().addItem(self.tempRubberBand)
		self.redrawActions()

	def lineClose(self):
		self.lineClosed = True
		self.capturedPoints.append(self.capturedPoints[0])
		self.redrawRubberBand()
		self.redrawTempRubberBand()
		self.redrawAreas()
		self.redrawActions()

	def lineOpen(self):
		self.lineClosed = False
		del self.capturedPoints[-1]
		self.redrawRubberBand()
		self.redrawTempRubberBand()
		self.redrawAreas()
		self.redrawActions()

	def showVertices(self):
		for i in range(len(self.capturedPoints)):
			vertexCoords = self.toCanvasCoordinates(self.toMapCoordinates(self.layer, self.capturedPoints[i]))
			if i == self.movingVertex:
				vertex = self.scene.addRect(vertexCoords.x() - 5, vertexCoords.y() - 5, 10, 10, QPen(QColor("green")), QBrush(QColor("green")))
				self.vertices.append(vertex)
			elif i == len(self.capturedPoints) - 1 and self.movingVertex == 0 and self.lineClosed:
				vertex = self.scene.addRect(vertexCoords.x() - 5, vertexCoords.y() - 5, 10, 10, QPen(QColor("green")), QBrush(QColor("green")))
				self.vertices.append(vertex)
			else:
				vertex = self.scene.addRect(vertexCoords.x() - 4, vertexCoords.y() - 4, 8, 8, QPen(QColor("red")), QBrush(QColor("red")))
				self.vertices.append(vertex)

	def deleteVertices(self):
		for i in range(len(self.vertices)):
			self.scene.removeItem(self.vertices[i])
		self.vertices = []

	def redrawVertices(self):
		self.deleteVertices()
		self.showVertices()

	def redrawActions(self):
		self.redrawActionMoveVertices()
		self.redrawActionLineClose()
		self.redrawActionLineOpen()
		self.redrawActionMoveLine()

	def redrawActionMoveVertices(self):
		self.actionMoveVertices.setEnabled(False)
		if len(self.capturedPoints) > 0:
			self.actionMoveVertices.setEnabled(True)

	def redrawActionLineClose(self):
		self.actionLineClose.setEnabled(False)
		if not self.lineClosed and len(self.capturedPoints) >= 3:
			self.actionLineClose.setEnabled(True)

	def redrawActionLineOpen(self):
		self.actionLineOpen.setEnabled(False)
		if self.lineClosed:
			self.actionLineOpen.setEnabled(True)

	def redrawActionMoveLine(self):
		self.actionMoveLine.setEnabled(False)
		if len(self.capturedPoints) > 0:
			self.actionMoveLine.setEnabled(True)